/*
 * DatalogTask.cpp
 *
 *  Created on: 2 Sep 2020
 *      Author: Jack
 */

#include <Tasks/DatalogTask.hpp>
#include <Wrappers/UDMAArray.hpp>
#include <Model/FirCoeffs.hpp>
#include "inc/hw_memmap.h"
#include "inc/tm4c123fh6pm.h"
#include "driverlib/adc.h"
#include "driverlib/sysctl.h"
#include "driverlib/interrupt.h"
#include "driverlib/udma.h"
#include "inc/hw_udma.h"
#include "utils/ustdlib.h"


#define TORQUE_ADC_PERIPH     SYSCTL_PERIPH_ADC0
#define TORQUE_ADC            ADC0_BASE
#define TORQUE_ADC_CHANNEL    ADC_CTL_CH7
#define TORQUE_ADC_INT        INT_ADC0SS3

Semaphore DatalogTask::gotReadingSemaphore(1,0);

tDMAControlTable DatalogTask::DMATaskList[4];
unsigned char DatalogTask::nextBuffer = 2;

DatalogTask::DatalogTask(Status* status,
                         QEI* motor_enc,
                         QEI* sensor_enc,
                         Settings* settings,
                         UART* uart)
:status(status),
 displ_enc(motor_enc),
 sensor_enc(sensor_enc),
 settings(settings),
 uart(uart)
{
    priority = 10;
    stackSize = 128;
    taskName = "DatalogTask";
}


void DatalogTask::Setup()
{
    // The ADC0 peripheral must be enabled for configuration and use.
    SysCtlPeripheralEnable(TORQUE_ADC_PERIPH);
    while (!SysCtlPeripheralReady(TORQUE_ADC_PERIPH)) ;

    SysCtlPeripheralEnable(SYSCTL_PERIPH_UDMA);
    while (!SysCtlPeripheralReady(SYSCTL_PERIPH_UDMA)) ;

    uDMAEnable();
    uDMAControlBaseSet(udmaArray);
    uDMAChannelAttributeEnable(UDMA_CHANNEL_ADC0, UDMA_ATTR_USEBURST);

    uDMAChannelControlSet(UDMA_CHANNEL_ADC0 | UDMA_PRI_SELECT, UDMA_SIZE_16 | UDMA_SRC_INC_NONE | UDMA_DST_INC_16 | UDMA_ARB_4);
    uDMAChannelTransferSet(UDMA_CHANNEL_ADC0 | UDMA_PRI_SELECT, UDMA_MODE_PINGPONG, (void *)&ADC0_SSFIFO0_R, &ADCbuffers[0][0], 250);
    uDMAChannelControlSet(UDMA_CHANNEL_ADC0 | UDMA_ALT_SELECT, UDMA_SIZE_16 | UDMA_SRC_INC_NONE | UDMA_DST_INC_16 | UDMA_ARB_4);
    uDMAChannelTransferSet(UDMA_CHANNEL_ADC0 | UDMA_ALT_SELECT, UDMA_MODE_PINGPONG, (void *)&ADC0_SSFIFO0_R, &ADCbuffers[1][0], 250);

    uDMAChannelEnable(UDMA_CHANNEL_ADC0 | UDMA_PRI_SELECT);
    uDMAChannelEnable(UDMA_CHANNEL_ADC0 | UDMA_ALT_SELECT);
/*


    uDMAChannelScatterGatherSet(UDMA_CHANNEL_ADC0, 4, DMATaskList, true);


    //copy tasks for buffers
    DMATaskList[0] = uDMATaskStructEntry(
            250,
            UDMA_SIZE_16,
            UDMA_SRC_INC_NONE,
            &ADC0_SSFIFO0_R,
            UDMA_DST_INC_16,
            &ADCbuffers[0][0],
            UDMA_ARB_4,
            UDMA_MODE_PER_SCATTER_GATHER);
    DMATaskList[1] = uDMATaskStructEntry(
            250,
            UDMA_SIZE_16,
            UDMA_SRC_INC_NONE,
            &ADC0_SSFIFO0_R,
            UDMA_DST_INC_16,
            &ADCbuffers[1][0],
            UDMA_ARB_4,
            UDMA_MODE_PER_SCATTER_GATHER);
    DMATaskList[2] = uDMATaskStructEntry(
            250,
            UDMA_SIZE_16,
            UDMA_SRC_INC_NONE,
            &ADC0_SSFIFO0_R,
            UDMA_DST_INC_16,
            &ADCbuffers[2][0],
            UDMA_ARB_4,
            UDMA_MODE_PER_SCATTER_GATHER);
    DMATaskList[3] = uDMATaskStructEntry(
            250,
            UDMA_SIZE_16,
            UDMA_SRC_INC_NONE,
            &ADC0_SSFIFO0_R,
            UDMA_DST_INC_16,
            &ADCbuffers[3][0],
            UDMA_ARB_4,
            UDMA_MODE_PER_SCATTER_GATHER);

*/
    ADCSequenceConfigure(TORQUE_ADC, 0, ADC_TRIGGER_ALWAYS, 0);
    ADCHardwareOversampleConfigure(TORQUE_ADC, 4);


    ADCSequenceStepConfigure(TORQUE_ADC, 0, 0, TORQUE_ADC_CHANNEL);
    ADCSequenceStepConfigure(TORQUE_ADC, 0, 1, TORQUE_ADC_CHANNEL);
    ADCSequenceStepConfigure(TORQUE_ADC, 0, 2, TORQUE_ADC_CHANNEL);
    ADCSequenceStepConfigure(TORQUE_ADC, 0, 3, TORQUE_ADC_CHANNEL | ADC_CTL_IE);
    ADCSequenceStepConfigure(TORQUE_ADC, 0, 4, TORQUE_ADC_CHANNEL);
    ADCSequenceStepConfigure(TORQUE_ADC, 0, 5, TORQUE_ADC_CHANNEL);
    ADCSequenceStepConfigure(TORQUE_ADC, 0, 6, TORQUE_ADC_CHANNEL);
    ADCSequenceStepConfigure(TORQUE_ADC, 0, 7, TORQUE_ADC_CHANNEL | ADC_CTL_IE | ADC_CTL_END);

    //
    // Since sample sequence 3 is now configured, it must be enabled.
    ADCSequenceEnable(TORQUE_ADC, 0);
    ADCSequenceDMAEnable(TORQUE_ADC, 0);
    //
    // Register the interrupt handler
    ADCIntRegister (TORQUE_ADC, 0, ISRHandler);


    //
    // Enable interrupts for ADC0 sequence 3 (clears any outstanding interrupts)
    IntPrioritySet(INT_ADC0SS0, (5 << 5) );
    ADCIntEnable(TORQUE_ADC, 0);
}

void DatalogTask::ISRHandler()
{
    uint32_t ints = ADCIntStatusEx(TORQUE_ADC, 0xffffffff);
    ADCIntClearEx(TORQUE_ADC, ints & ADC_INT_SS0);

//    uDMAChannelScatterGatherSet(UDMA_CHANNEL_ADC0, 4, DMATaskList, true);
//    uDMAChannelEnable(UDMA_CHANNEL_ADC0 | UDMA_PRI_SELECT);

    ADCSequenceDMAEnable(TORQUE_ADC, 0);
    gotReadingSemaphore.GiveFromISR();
}

void DatalogTask::TaskMethod()
{
    TickType_t xLastWakeTime = xTaskGetTickCount();

    volatile UBaseType_t uxHighWaterMark;

    while(1)
    {
        uxHighWaterMark = uxTaskGetStackHighWaterMark( NULL );


        //wait for samples
        gotReadingSemaphore.Take();

        //reset pingpong for next buffer
        uint32_t channel = UDMA_CHANNEL_ADC0 | ((nextBuffer % 2 == 0) ? UDMA_PRI_SELECT : UDMA_ALT_SELECT);
        uDMAChannelControlSet(channel, UDMA_SIZE_16 | UDMA_SRC_INC_NONE | UDMA_DST_INC_16 | UDMA_ARB_4);
        uDMAChannelTransferSet(channel, UDMA_MODE_PINGPONG, (void *)&ADC0_SSFIFO0_R, &ADCbuffers[nextBuffer % 4][0], 250);

        uDMAChannelEnable(channel);
        nextBuffer++;

        uint32_t ulValue  = filteredSample();

        int32_t rawSignedValue = ulValue - 0x80000000;//underflow negative values, correct zero-offset, then reinterpret as signed.

        int64_t longlongSignedValue = rawSignedValue;

        //scale to data range (-4/5 to 4/5 of full signed range (or 1/10 to 9/10 of unsigned range)) use larger data type to avoid overflows
        longlongSignedValue = (longlongSignedValue * 5)/4;

        // cal = m_c*uncal + a_c

        int multiplicative_calibration = 10157; // (1.0157 multiplied by 10000)
        int additive_calibration = 44517336; // 0.02073 of full range (-2^31:+2^31)

        int bin_index = (sensor_enc->getPos() / CAL_BIN_SIZE) % CAL_BINS;
        //TODO: add linear interpolation to calibration correction

        int64_t longlongCalibratedValue = (longlongSignedValue * multiplicative_calibration)/10000 + additive_calibration - calTable[bin_index];

        //scale to torque range (with 2 dp decimal fixed-point type) (-5(-500) to 5(500) Nm over data range)
        int16_t torqueFPValue2dp = (longlongCalibratedValue*500)/0x80000000/*2^31*/;

        //scale to torque range (with 3 dp decimal fixed-point type) (-5(-5000) to 5(5000) Nm over data range)
        int16_t torqueFPValue3dp = (longlongCalibratedValue*5000)/0x80000000/*2^31*/;


        status->setSensorEncoderRaw(sensor_enc->getPos());
        status->setSensorTorque(torqueFPValue3dp);
        status->setDisplacmentEncoderRaw(displ_enc->getPos());

        //prevent accidental calibration under load
        if(abs(torqueFPValue2dp) > 100 /*1Nm*/)
        {
            settings->endSelfCal();
            for(int i = 0; i<CAL_BINS; i++)
            {
                calAccumulator[i] = 0;
                calCount[i] = 0;
            }
        }

        if(settings->getSelfCalActive())
        {

            calAccumulator[bin_index] += (longlongCalibratedValue + calTable[bin_index]);
            calCount[bin_index]++;
            if(calCount[bin_index] > CAL_MAX_SAMPLES)
            {
                settings->endSelfCal();
                for(int i = 0; i<CAL_BINS; i++)
                {
                    calTable[i] = calAccumulator[i]/calCount[i];
                    calAccumulator[i] = 0;
                    calCount[i] = 0;
                }
            }
        }

        if(settings->getOutputSwitchTargetRaw() == false)
        {
            if(settings->getOutputSwitchPend())
            {
                settings->setOutputSwitchPend(false);
            }
            continue;
        }
        else if(!settings->getOutputSwitchPend())
        {
            int invert = settings->getInvert() ? -1 : 1;
            char string[60];
            usnprintf (string, sizeof(string), "T:%d;R:%d;D:%d;\r\n",
                       status->getSensorTorque()*invert, status->getSensorEncoderCal()*invert, status->getDisplacmentEncoderCal());
            uart->SendSync(string);
            if(status->newOutputString())
            {
                const char* string = status->getOutputString();
                uart->SendSync("!");
                uart->SendSync(string);
                uart->SendSync("\r\n");
            }
        }

        //vTaskDelayUntil(&xLastWakeTime, 1000);

    }
}

uint32_t DatalogTask::filteredSample()
{
    uint64_t register accumulator = 0;

    unsigned char nextBuffer = this->nextBuffer;//copy locally to avoid unexpected changes

    //last part of first buffer with first part of forward filter
    uint16_t* bufferPtr = &ADCbuffers[(nextBuffer + 3) % 4][123];

    const register uint16_t* coeffsPtr = &firCoeffs[0];
    const register uint16_t* coeffsEnd = &firCoeffs[126];

    while(coeffsPtr <= coeffsEnd)
    {
        accumulator = accumulator + (((uint32_t) *coeffsPtr) * ((uint32_t) *bufferPtr));
        bufferPtr++;
        coeffsPtr++;
    }


    //first part of second buffer with second part of forward filter
    bufferPtr = &ADCbuffers[(nextBuffer) % 4][0];

    coeffsEnd = &firCoeffs[313];

    while(coeffsPtr <= coeffsEnd)
    {
        accumulator = accumulator + (((uint32_t) *coeffsPtr) * ((uint32_t) *bufferPtr));
        bufferPtr++;
        coeffsPtr++;
    }

    //second part of second buffer with first part of reversed filter

    coeffsPtr = &firCoeffs[312];
    coeffsEnd = &firCoeffs[250];


    while(coeffsPtr >= coeffsEnd)
    {
        accumulator = accumulator + (((uint32_t) *coeffsPtr) * ((uint32_t) *bufferPtr));
        bufferPtr++;
        coeffsPtr--;
    }
    //third buffer with second part of reversed filter
    bufferPtr = &ADCbuffers[(nextBuffer + 1) % 4][0];

    coeffsEnd = &firCoeffs[0];

    while(coeffsPtr >= coeffsEnd)
    {
        accumulator = accumulator + (((uint32_t) *coeffsPtr) * ((uint32_t) *bufferPtr));
        bufferPtr++;
        coeffsPtr--;
    }

    //scale down accumulator to fit almost exactly in 32 bits.
    //normal
    accumulator = (accumulator * 1000)/18401;//divide by 18.401
    //delta (one 65535)
    //accumulator = (accumulator * 10000)/625; //divide by 0.062
    //average (all 65535)
    //accumulator = (accumulator * 1000 * 65535)/39187; //divide by 39.187

    return accumulator;
}
