/*
 * MotorTask.cpp
 *
 *  Created on: 9 Oct 2020
 *      Author: Jack
 */

#include <Tasks/MotorTask.hpp>
#include "driverlib/sysctl.h"
#include "driverlib/timer.h"
#include "driverlib/pwm.h"
#include "driverlib/gpio.h"
#include "inc/hw_memmap.h"
#include "inc/hw_types.h"
#include "driverlib/pin_map.h"
#include "inc/tm4c123gh6pm.h"

#include "FreeRTOSConfig.h"
#define SYS_CLK configCPU_CLOCK_HZ
#define PULSE_MULT 2
#define CPS_MAX 80000

void MotorTask::Setup()
{
    motor_ena->setActiveState(1);
    motor_ena->setActiveState(0);
#ifdef PWM_BASED
    //setup wide timer 2(B) pwm 1 (PD1)
    //48 bit down-count pwm
    //gpio
    SysCtlPeripheralEnable(SYSCTL_PERIPH_GPIOD);
    while(!SysCtlPeripheralReady(SYSCTL_PERIPH_GPIOD));

    GPIOPinConfigure(GPIO_PD1_WT2CCP1);
    GPIOPinTypeGPIOOutput(GPIO_PORTD_BASE, GPIO_PIN_1);
    GPIOPinTypeTimer(GPIO_PORTD_BASE, GPIO_PIN_1);

    //pwm(timer)
    SysCtlPeripheralEnable(SYSCTL_PERIPH_WTIMER2);
    while(!SysCtlPeripheralReady(SYSCTL_PERIPH_WTIMER2));

    TimerConfigure(WTIMER2_BASE, TIMER_CFG_SPLIT_PAIR | TIMER_CFG_B_PWM);
    TimerUpdateMode(WTIMER2_BASE, TIMER_B, TIMER_UP_LOAD_TIMEOUT | TIMER_UP_MATCH_TIMEOUT);


    //setup counter for PB7 (timer0(B) ccp1)
    //24 bit up-counter
    //counter(timer)
    SysCtlPeripheralEnable(SYSCTL_PERIPH_TIMER0);
    while(!SysCtlPeripheralReady(SYSCTL_PERIPH_TIMER0));

    TimerConfigure(TIMER0_BASE, TIMER_CFG_SPLIT_PAIR | TIMER_CFG_B_CAP_COUNT_UP);
    TimerEnable(TIMER0_BASE, TIMER_B);

    //gpio
    SysCtlPeripheralEnable(SYSCTL_PERIPH_GPIOB);
    while(!SysCtlPeripheralReady(SYSCTL_PERIPH_GPIOB));

    GPIOPinConfigure(GPIO_PB7_T0CCP1);
    GPIOPinTypeTimer(GPIO_PORTB_BASE, GPIO_PIN_7);


    //setup high precision wtimer0(BOTH)
    //64 bit up-counter (80MHz = 10000years runtime before overflow)
    SysCtlPeripheralEnable(SYSCTL_PERIPH_WTIMER0);
    while(!SysCtlPeripheralReady(SYSCTL_PERIPH_WTIMER0));
    TimerConfigure(WTIMER0_BASE, TIMER_CFG_PERIODIC_UP);
    TimerEnable(WTIMER0_BASE, TIMER_BOTH);
#endif

}

void MotorTask::TaskMethod()
{

    TickType_t xLastWakeTime = xTaskGetTickCount();

    bool active = false;
    int tickIndex = 0;
    MotionSegment activeMotion;

#ifdef PWM_BASED
    int currspeed = 0;
    uint32_t pwm_load_value_old = 0xFFFFFFFF;
    uint32_t pwm_compare_value_old = 0xFFFFFFFF;
    int64_t long_pulse_accumulator = 0;
    int64_t long_pulses_expected = 0;
    uint64_t t0_timestamp = 0;
    uint64_t old_timestamp = 0;
    uint16_t old_pulse_count = 0;
    int error = 0;
#endif

    volatile UBaseType_t uxHighWaterMark;
    while(1)
    {
        uxHighWaterMark = uxTaskGetStackHighWaterMark( NULL );

        //calculate required steps this for this ms

        int pulses = 0;

#ifdef PWM_BASED

        uint16_t new_pulse_count = TimerValueGet(TIMER0_BASE, TIMER_B);
        uint64_t new_timestamp = TimerValueGet64(WTIMER0_BASE);
#endif

        if(!active)
        {
            tickIndex = 0;

#ifdef PWM_BASED
            t0_timestamp = new_timestamp;
#endif
            active = motionQueue->Dequeue(&activeMotion);
            status->setMotorSpeed(0);
        }

        if(active)
        {
            pulses = activeMotion.pulsesForTick(tickIndex)*PULSE_MULT;
            status->setMotorSpeed(activeMotion.tickPPS(tickIndex));
            tickIndex++;
            if(tickIndex == activeMotion.getTickLength())
            {
                active = false;
            }
        }

        //check if motor enabled
        bool enabled = settings->getEnableMotor();
        motor_ena->setActiveState(enabled);
        if(!enabled)
        {
#ifdef PWM_BASED
            long_pulse_accumulator = 0;
            long_pulses_expected = 0;
#endif
            pulses = 0;
            active = false;
            tickIndex = 0;
            while(motionQueue->Dequeue(&activeMotion));//dequeue until empty
            //TimerDisable(WTIMER2_BASE, TIMER_B);
            //TimerMatchSet(WTIMER2_BASE, TIMER_B, 0);
            //TimerLoadSet(WTIMER2_BASE, TIMER_B, 0);
        }

#ifdef PWM_BASED
        //update frequency


        int cps = 0;
        if(active || tickIndex != 0)
        {
            cps = abs(activeMotion.tickPPS(tickIndex));
            if(cps > CPS_MAX)
            {
                cps = CPS_MAX;
            }
        }

        //calculate positional correction
        //TODO
        bool curdir = motor_dir->getActiveState() != settings->getInvert();
        bool newdir = activeMotion.tickPPS(tickIndex) > 0;



        long_pulse_accumulator += (curdir ? 1 : -1)*(new_pulse_count - old_pulse_count);//intentional use of 32-bit overflow

        if((new_timestamp+100 - old_timestamp)%80000 < 200)
        {
            error = long_pulse_accumulator - long_pulses_expected;
        }

        const int P_Coeff =1;
        //cps += P_Coeff*error; correction not surrently working

        old_pulse_count = new_pulse_count;
        long_pulses_expected += pulses;//accounts for direction already
        if(cps < 0)
        {
            cps = -cps;
            newdir = !newdir;
        }

        //update pwm
        if(cps != 0)
        {
            const int margin = 100;
            motor_dir->setActiveState(newdir != settings->getInvert());
            uint32_t required_cylces = (SYS_CLK + ((cps*PULSE_MULT)/2))/(cps*PULSE_MULT); //accurately rounding integer division

            //for pwm setup

            uint32_t pwm_load_value = required_cylces-1;//datasheet for motor driver isn't super clear that this is what is needed
            uint32_t pwm_compare_value = (pwm_load_value + 1)/2;

            volatile uint32_t pwm_curr_value = TimerValueGet(WTIMER2_BASE, TIMER_B);
            bool should_scale = (pwm_load_value*2 < pwm_load_value_old && pwm_curr_value > margin) || pwm_load_value_old == 0xFFFFFFFF;

            if(should_scale)
            {
                TimerDisable(WTIMER2_BASE, TIMER_B);
                pwm_curr_value = TimerValueGet(WTIMER2_BASE, TIMER_B);
            }


            uint32_t pwm_new_scaled_value = (((uint64_t)pwm_curr_value)*pwm_load_value_old)/pwm_load_value;
            if(pwm_new_scaled_value < margin)
            {
                pwm_new_scaled_value = margin;
            }

            TimerMatchSet(WTIMER2_BASE, TIMER_B, pwm_compare_value);
            if(should_scale)
            {
                TimerLoadSet(WTIMER2_BASE, TIMER_B, pwm_new_scaled_value);
                TimerEnable(WTIMER2_BASE, TIMER_B);
            }
            TimerLoadSet(WTIMER2_BASE, TIMER_B, pwm_load_value);

            pwm_load_value_old = pwm_load_value;
            pwm_compare_value_old = pwm_compare_value;

        }
        else
        {
            TimerDisable(WTIMER2_BASE, TIMER_B);
            pwm_load_value_old = 0xFFFFFFFF;
            pwm_compare_value_old = 0xFFFFFFFF;
            TimerMatchSet(WTIMER2_BASE, TIMER_B, 0);
            TimerLoadSet(WTIMER2_BASE, TIMER_B, 0);
        }
        old_timestamp = new_timestamp;
#endif

#ifndef PWM_BASED
        if(pulses != 0)
        {

            motor_dir->setActiveState((pulses > 0) != settings->getInvert());
            int absPulses = abs(pulses)*PULSE_MULT;
            for(int i = 0; i < absPulses ; i++)
            {
                SysCtlDelay(SysCtlClockGet() / 3 / 400000); //2.5s delay
                motor_pul->setActiveState(1);
                SysCtlDelay(SysCtlClockGet() / 3 / 400000); //2.5s delay
                motor_pul->setActiveState(0);
            }
        }
#endif

        vTaskDelayUntil(&xLastWakeTime,  10);
    }
}

MotorTask::MotorTask(GPIO* motor_pul, GPIO* motor_dir, GPIO* motor_ena, GPIO* motor_alm, GPIO* motor_pend,Queue<MotionSegment>* motionQueue, Status* status, Settings* settings)
:motor_pul(motor_pul),motor_dir(motor_dir),motor_ena(motor_ena),motor_alm(motor_alm),motor_pend(motor_pend),motionQueue(motionQueue), status(status), settings(settings)
{
    priority = 12;
    stackSize = 128;
    taskName = "MotorTask";
}
